% FLENG compiler - clause indexing

compile_group(N/A, I, EL, G, Lc1, Lc) :-
    gen([define(A, I, EL, N), reset]),
    partition_on_first_arg(G, ACs, LCs, TCs, OCs),
    select_indexing_method(ACs, LCs, TCs, OCs, G, Lc1, Lc2),
    label(Lc2, Lc, Lf),
    gen([not_suspending(Lf), suspend, label(Lf)]),
    gen([fail(A)]).

select_indexing_method(ACs, LCs, TCs, OCs, _, Lc1, Lc) :-
    length(ACs, Len1),
    length(LCs, Len2),
    length(TCs, Len3),
    Len1 + Len2 + Len3 > 2,
    !,
    compile_indexed(ACs, LCs, TCs, OCs, Lc1, Lc).
select_indexing_method(_, _, _, _, All, Lc1, Lc) :-
    compile_unindexed(All, none, yes, Lc1, Lc).

partition_on_first_arg([], [], [], [], []).
partition_on_first_arg([t((H :- T), Vs, Ss)|G], ACs, LCs, TCs, OCs) :-
    H =.. [_, Arg|_],
    partition_on_first_arg(Arg, t((H :-T), Vs, Ss), G, ACs, LCs, TCs, OCs).
partition_on_first_arg(G, [], [], [], G).

partition_on_first_arg('$VAR'(_), C, G, ACs, LCs, TCs, [C|OCs]) :-
    partition_on_first_arg(G, ACs, LCs, TCs, OCs).
partition_on_first_arg('!'(Arg), C, G, ACs, LCs, TCs, OCs) :-
    partition_on_first_arg(Arg, C, G, ACs, LCs, TCs, OCs).
partition_on_first_arg(X, C, G, [C|ACs], LCs, TCs, OCs) :-
    atom(X),
    partition_on_first_arg(G, ACs, LCs, TCs, OCs).
partition_on_first_arg(X, C, G, [C|ACs], LCs, TCs, OCs) :-
    integer(X),
    X >= -16384, X =< 16383,
    partition_on_first_arg(G, ACs, LCs, TCs, OCs).
partition_on_first_arg([_|_], C, G, ACs, [C|LCs], TCs, OCs) :-
    partition_on_first_arg(G, ACs, LCs, TCs, OCs).
partition_on_first_arg(F, C, G, ACs, LCs, [C|TCs], OCs) :-
    compound(F),
    partition_on_first_arg(G, ACs, LCs, TCs, OCs).
partition_on_first_arg(_, C, G, ACs, LCs, TCs, [C|OCs]) :-
    partition_on_first_arg(G, ACs, LCs, TCs, OCs).    

compile_unindexed([], Fail, _, Lc, Lc) :-
    emit_fail_label(Fail).
compile_unindexed([t(Term, _, Ss)|G], Fail, ForceArg1, Lc1, Lc) :-
    emit_fail_label(Fail),
    label(Lc1, Lc2, Fail2),
    compile_clause(Term, Ss, ForceArg1, Fail2, Lc2, Lc3),
    compile_unindexed(G, Fail2, yes, Lc3, Lc).

emit_fail_label(none).
emit_fail_label(Fail) :-
    number(Fail), 
    gen(try_else(Fail)).

% force first arg and dispatch according to type

compile_indexed(ACs, LCs, TCs, OCs, Lc1, Lc) :-
    label(Lc1, Lc2, Lo),
    gen(force_arg(Lo)),
    compile_indexed2(ACs, LCs, TCs, OCs, none, Lo, Lc2, Lc).

compile_indexed2([C|ACs], LCs, TCs, OCs, Fail, Lo, Lc1, Lc) :-
    emit_fail_label(Fail),
    label(Lc1, Lc2, Fail2),
    gen([getarg(0), switch_atomic(Fail2)]),
    compile_lookup_table([C|ACs], Fail2, Lc2, Lc3),
    compile_indexed2([], LCs, TCs, OCs, Fail2, Lo, Lc3, Lc).
compile_indexed2([], [C|LCs], TCs, OCs, Fail, Lo, Lc1, Lc) :-
    emit_fail_label(Fail),
    label(Lc1, Lc2, Fail2),
    gen([getarg(0), switch_list(Fail2)]),
    compile_unindexed([C|LCs], none, no, Lc2, Lc3),
    compile_indexed2([], [], TCs, OCs, Fail2, Lo, Lc3, Lc).
compile_indexed2([], [], [C|TCs], OCs, Fail, Lo, Lc1, Lc) :-
    emit_fail_label(Fail),
    gen([getarg(0), switch_tuple(Lo)]),
    compile_unindexed([C|TCs], none, no, Lc1, Lc2),
    compile_indexed2([], [], [], OCs, none, Lo, Lc2, Lc).
compile_indexed2([], [], [], OCs, Fail, Lo, Lc1, Lc) :-
    emit_fail_label(Fail),
    gen(label(Lo)),
    compile_unindexed(OCs, none, no, Lc1, Lc).

compile_lookup_table(ACs, Fail, Lc1, Lc) :-
    length(ACs, N),
    N > 2,
    labels(N, Lc1, Lc2, Lfs),
    head_args(ACs, Args),
    gen(lookup(N, Args, Lfs, Fail)),
    compile_lookup_table2(ACs, Lfs, Fail, Lc2, Lc).
compile_lookup_table(ACs, _, Lc1, Lc) :-
    compile_unindexed(ACs, none, no, Lc1, Lc).

head_args([], []).
head_args([t((H :- _), _, _)|Cs], [Arg|Args]) :-
    H =.. [_, Arg|_],
    head_args(Cs, Args).

compile_lookup_table2([], _, _, Lc, Lc).
compile_lookup_table2([t(Term, _, Ss)|Cs], [L|Lfs], Fail, Lc1, Lc) :-
    gen(label(L)),
    compile_clause(Term, Ss, no, Fail, Lc1, Lc2),
    compile_lookup_table2(Cs, Lfs, Fail, Lc2, Lc).

labels(0, Lc, Lc, []).
labels(N, Lc1, Lc2, [L|Ls]) :-
    label(Lc1, Lc3, L),
    N2 is N - 1,
    labels(N2, Lc3, Lc2, Ls).
