% FLENG compiler - peephole optimization

init_optimizer :- 
    g_assign(pending, []).

gen([]) :- !.
gen([X|R]) :-
    !,
    g_read(output, Out),
    g_read(pending, P),
    append(P, [X|R], All),
    optimize(All, Emit, Remaining),
    g_assign(pending, Remaining), 
    asm(Emit, Out).
gen(X) :- gen([X]).

finalize_optimizer :-
    g_read(pending, Code),
    g_read(output, Out),
    asm(Code, Out).

optimize([], [], []) :- !.
optimize(Code, Emit, Remain) :-
    opt(Pat, Prod, Id),
    match_opt_pattern(Code, Pat, Code, Prod, Id, Emit, Remain),
    !.
optimize([Asm|Code], [Asm|Rest], Remain) :-
    optimize(Code, Rest, Remain).

match_opt_pattern(Code, [], All, Prod, Id, Emit, []) :-
    append(Prod, Code, Emit),
    debug_compiler(opt, Id, All, Emit).
match_opt_pattern([], [_|_], All, _, _, [], All).
match_opt_pattern([Instr|Code], [Instr|Pat], All, Prod, Id, Emit, Remain) :-
    match_opt_pattern(Code, Pat, All, Prod, Id, Emit, Remain).

% optimization rules

opt([try_else(L1), try_else(L2)], [label(L1), try_else(L2)], 1).
opt([getenv(I), putparam(J)], [putparam_env(J, I)], 2).
opt([getenv(I), deref, putparam(J)], [putparam_env(J, I)], 3).
opt([drop(I), drop(J)], [drop2(I, J)], 4) :- target_architecture(uxn).
opt([check_integer, intval], [intval_checked], 5).
opt([mkint(N), intval], [mkint_val(N)], 6).
opt([putarg(I), addrefx], [storearg(I)], 7).
